#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef long INT32;
typedef short INT16;
typedef char INT8;
typedef unsigned long UINT32;
typedef unsigned short UINT16;
typedef unsigned char UINT8;

#define FIND_ANY 0
#define FIND_TYPE 0x0001
#define FIND_SUBTYPE 0x0002
#define FIND_GROUP 0x0004
#define FIND_INSTANCE 0x0008

typedef struct tagFindRec {
    UINT32 type;
    UINT32 subtype;
    UINT32 group;
    UINT32 instance;
    UINT32 size;
    UINT32 index;
} FindRec, *pFindRec;

void loadtoc(FILE *f, INT8 *s);
UINT32 find2(UINT32 cmd, pFindRec pFind);
void do_list(FILE *f, int gb, UINT32 g, int tb, UINT32 t);
int opt_parse(INT8 *o, int x);
UINT32 hex2int(INT8 *s);

UINT32 gSize, gVer, *gToc;
INT8 gName[4096];

void loadtoc(FILE *f, INT8 *s)
{
	int status = 0;
	UINT32 xStart, xSize, xVer, *pToc;

	gToc = NULL;
	if (f != NULL) {
		fseek(f, 0x28, 0);
		fread(&xStart, 4, 1, f);
		fread(&xSize, 4, 1, f);
		fseek(f, 0x3C, 0);
		fread(&xVer, 4, 1, f);              // subversion (1/2)

		if (xSize != 0) {
			fseek(f, xStart, 0);
			pToc = (UINT32 *) malloc(xSize);
			if (pToc != NULL) {
				fread(pToc, xSize, 1, f);   // load toc
				status = 1;
			}
		}

		if (status) {
			gToc = pToc;
			gSize = xSize;
			gVer = xVer;
			strcpy(gName, s);
		}
	}
}

UINT32 find2(UINT32 cmd, pFindRec pFind)
{
    UINT32 s, v, x, *t;
    UINT32 eType, eSubType, eGroup, eInstance, eOffset, eSize;

    s = (gSize >> 2);
    v = gVer;
    t = gToc;
    x = pFind->index;
    eSubType = 0;

    while (x < s) {
        eType = t[x++];
        eGroup = t[x++];
        eInstance = t[x++];
        if (v > 1) eSubType = t[x++];
        eOffset = t[x++];
        eSize = t[x++];

        if (!( ((cmd & FIND_TYPE) && (eType != pFind->type)) ||
               ((cmd & FIND_SUBTYPE) && (eSubType != pFind->subtype)) ||
               ((cmd & FIND_GROUP) && (eGroup != pFind->group)) ||
               ((cmd & FIND_INSTANCE) && (eInstance != pFind->instance))
           ) ) {
                 pFind->type = eType;
                 pFind->subtype = eSubType;
                 pFind->group = eGroup;
                 pFind->instance = eInstance;
                 pFind->size = eSize;
                 pFind->index = x;
                 return(eOffset);
        }
    }
    return(0);
}

void do_list(FILE *f, int gb, UINT32 g, int tb, UINT32 t)
{
	UINT32 o, ft;
	FindRec fr = { 0, 0, 0, 0, 0, 0 };

	ft = FIND_ANY;
	if (gb != 0) {
		ft += FIND_GROUP;
		fr.group = g;
	}
	if (tb != 0) {
		ft += FIND_TYPE;
		fr.type = t;
	}
	o = find2(ft, &fr);
	while (o != 0) {
		fprintf(f, "Group = 0x%X, Type = 0x%X, Instance = 0x%X, Subtype = 0x%X\n", \
			fr.group, fr.type, fr.instance, fr.subtype);
		o = find2(ft, &fr);
	}
}

int opt_parse(INT8 *o, int x)
{
	if (stricmp(o, "-g") == 0) return 1;
	if (stricmp(o, "-t") == 0) return 2;
	if (x == 1) return 101;
	if (x == 2) return 102;
	return 0;
}

UINT32 hex2int(INT8 *s)
{
	INT8 *s2 = s;
	UINT32 v;
	int r;

	if (strlen(s) > 2) {
		if (stricmp(s, "0x") == 0) {
			s2 = s + 2;
		}
	}
	if (sscanf(s2, "%x", &v) == 1)
		return v;
	else
		return 0;
}

int main(int argc, char *argv[])
{
	FILE *outFile = stdout;
	FILE *packFile = NULL;
	int i, x = 0, pf = 0;
	int gb = 0, tb = 0;
	UINT32 g, t;

	if (argc > 1) {
		for (i = 1; i < argc; i++) {
			x = opt_parse(argv[i], x);
			if (x == 0) {
				if (pf == 0) {
					packFile = fopen(argv[i], "rb");
					pf = 1;
				} else if (outFile == stdout) {
					outFile = fopen(argv[i], "w");
					if (outFile == NULL) outFile = stdout;
				}
			}
			if (x == 101) {
				g = hex2int(argv[i]);
				gb = 1;
			}
			if (x == 102) {
				t = hex2int(argv[i]);
				tb = 1;
			}
		}
		loadtoc(packFile, argv[1]);
		if (gToc != NULL) {
			do_list(outFile, gb, g, tb, t);
			free(gToc);
		}
	} else {
		printf("ls usage:\n\nls [-g group] [-t type] package [output]\n\n");
	}
	if (outFile != stdout) fclose(outFile);
	if (packFile != NULL) fclose(packFile);
	return 0;
}
